# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
The parallel master class is a metajob consisting of a list of jobs which are executed in parallel.
"""

from collections import OrderedDict
from datetime import datetime

import numpy as np
import pandas
import multiprocessing
import importlib
from pyiron_base.jobs.job.generic import GenericJob
from pyiron_base.jobs.job.core import _doc_str_job_core_args
from pyiron_base.jobs.master.generic import GenericMaster, _doc_str_generic_master_attr
from pyiron_base.jobs.master.submissionstatus import SubmissionStatus
from pyiron_base.jobs.job.extension.jobstatus import JobStatus
from pyiron_base.state import state
from pyiron_base.jobs.job.wrapper import job_wrapper_function
from pyiron_base.utils.deprecate import deprecate

__author__ = "Joerg Neugebauer, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


# Modular Docstrings
_doc_str_parallel_master_attr = (
    _doc_str_generic_master_attr
    + "\n"
    + """\
        .. attribute:: ref_job

            Reference job template from which all jobs within the ParallelMaster are generated.

        .. attribute:: number_jobs_total

            Total number of jobs
"""
)


class ParallelMaster(GenericMaster):
    __doc__ = (
        """
    MasterJob that handles the creation and analysis of several parallel jobs (including master and
    continuation jobs), Examples are Murnaghan or Phonon calculations.

    Subclasses *must* implement :meth:`.collect_output()`.  Additionally :attr:`._job_generator` must be
    initialized with an instance of :class:`.JobGenerator` in the subclasses' `__init__`.
"""
        + "\n"
        + _doc_str_job_core_args
        + "\n"
        + _doc_str_parallel_master_attr
    )

    def __init__(self, project, job_name):
        super(ParallelMaster, self).__init__(project, job_name=job_name)
        self.__version__ = "0.3"
        self._ref_job = None
        self._output = GenericOutput()
        self._job_generator = None
        self.submission_status = SubmissionStatus(db=project.db, job_id=self.job_id)
        self.refresh_submission_status()

    @property
    def ref_job(self):
        """
        Get the reference job template from which all jobs within the ParallelMaster are generated.

        Returns:
            GenericJob: reference job
        """
        if self._ref_job:
            return self._ref_job
        try:
            ref_job = self[0]
            if isinstance(ref_job, GenericJob):
                self._ref_job = ref_job
                self._ref_job._job_id = None
                self._ref_job._status = JobStatus(db=self.project.db)
                return self._ref_job
            else:
                return None
        except IndexError:
            return None

    @ref_job.setter
    def ref_job(self, ref_job):
        """
        Set the reference job template from which all jobs within the ParallelMaster are generated.

        Args:
            ref_job (GenericJob): reference job
        """
        self.append(ref_job)

    @property
    def number_jobs_total(self):
        """
        Get number of total jobs

        Returns:
            int: number of total jobs
        """
        return self.submission_status.total_jobs

    @number_jobs_total.setter
    def number_jobs_total(self, num_jobs):
        """
        Set number of total jobs (optional: default = None)

        Args:
            num_jobs (int): number of submitted jobs
        """
        self.submission_status.total_jobs = num_jobs

    def reset_job_id(self, job_id=None):
        """
        Reset the job id sets the job_id to None as well as all connected modules like JobStatus and SubmissionStatus.
        """
        super(ParallelMaster, self).reset_job_id(job_id=job_id)
        if job_id is not None:
            self.submission_status = SubmissionStatus(db=self.project.db, job_id=job_id)
        else:
            self.submission_status = SubmissionStatus(
                db=self.project.db, job_id=self.job_id
            )

    def collect_output(self):
        """
        Collect the output files of the external executable and store the information in the HDF5 file. This method has
        to be implemented in the individual meta jobs derived from the ParallelMaster.
        """
        raise NotImplementedError("Implement in derived class")

    def collect_logfiles(self):
        """
        Collect the log files of the external executable and store the information in the HDF5 file. This method is
        currently not implemented for the ParallelMaster.
        """
        pass

    def output_to_pandas(self, sort_by=None, h5_path="output"):
        """
        Convert output of all child jobs to a pandas Dataframe object.

        Args:
            sort_by (str): sort the output using pandas.DataFrame.sort_values(by=sort_by)
            h5_path (str): select child output to include - default='output'

        Returns:
            pandas.Dataframe: output as dataframe
        """
        # TODO: The output to pandas function should no longer be required
        with self.project_hdf5.open(h5_path) as hdf:
            for key in hdf.list_nodes():
                self._output[key] = hdf[key]
        df = pandas.DataFrame(self._output)
        if sort_by is not None:
            df = df.sort_values(by=sort_by)
        return df

    # TODO: make it more general and move it then into genericJob
    def show_hdf(self):
        """
        Display the output of the child jobs in a human readable print out
        """
        try:
            display = getattr(importlib.import_module("IPython"), "display")
        except ModuleNotFoundError:
            print("show_hdf() requires IPython to be installed.")
        else:
            for nn in self.project_hdf5.list_groups():
                with self.project_hdf5.open(nn) as hdf_dir:
                    display.display(nn)
                    if nn.strip() == "output":
                        display.display(self.output_to_pandas(h5_path=nn))
                        continue
                    for n in hdf_dir.list_groups():
                        display.display("-->" + n)
                        try:
                            display.display(hdf_dir.get_pandas(n))
                        except Exception as e:
                            print(e)
                            print("Not a pandas object")

    def save(self):
        """
        Save the object, by writing the content to the HDF5 file and storing an entry in the database.

        Returns:
            (int): Job ID stored in the database
        """
        job_id = super(ParallelMaster, self).save()
        self.refresh_submission_status()
        return job_id

    def refresh_submission_status(self):
        """
        Refresh the submission status - if a job ID job_id is set then the submission status is loaded from the
        database.
        """
        if self.job_id:
            self.submission_status = SubmissionStatus(
                db=self.project.db, job_id=self.job_id
            )
            self.submission_status.refresh()

    def interactive_ref_job_initialize(self):
        """
        To execute the reference job in interactive mode it is necessary to initialize it.
        """
        if len(self._job_name_lst) > 0:
            self._ref_job = self.pop(-1)
            job_name = self.job_name + "_" + self._ref_job.job_name
            self._ref_job.job_name = job_name
            self._ref_job.project_hdf5 = self.child_hdf(job_name)
            if self._job_id is not None and self._ref_job._master_id is None:
                self._ref_job.master_id = self.job_id

    def copy(self):
        """
        Copy the GenericJob object which links to the job and its HDF5 file

        Returns:
            :class:`.GenericJob`: New object pointing to the same job
        """
        new_job = super(ParallelMaster, self).copy()
        if self.ref_job is not None:
            new_job.ref_job = self.ref_job
        return new_job

    def _after_generic_copy_to(self, original, new_database_entry, reloaded):
        super()._after_generic_copy_to(original, new_database_entry, reloaded)
        self.submission_status = SubmissionStatus(
            db=self._hdf5.project.db, job_id=self.job_id
        )

    def is_finished(self):
        """
        Check if the ParallelMaster job is finished - by checking the job status and the submission status.

        Returns:
            bool: [True/False]
        """
        if self.status.finished:
            return True
        if len(self.child_ids) < len(self._job_generator):
            return False
        return set(
            [self.project.db.get_job_status(child_id) for child_id in self.child_ids]
        ) < {"finished", "busy", "refresh", "aborted", "not_converged"}

    def iter_jobs(self, convert_to_object=True):
        """
        Iterate over the jobs within the ListMaster

        Args:
            convert_to_object (bool): load the full GenericJob object (default) or just the HDF5 / JobCore object

        Returns:
            yield: Yield of GenericJob or JobCore
        """
        project_working_directory = self.project.open(self.job_name + "_hdf5")
        for job_id in self._get_jobs_sorted():
            yield project_working_directory.load(
                job_id, convert_to_object=convert_to_object
            )

    def _get_jobs_sorted(self):
        job_names = self.child_names.values()
        return [
            j
            for j in [
                self._job_generator.job_name(p)
                for p in self._job_generator.parameter_list
            ]
            if j in job_names
        ]

    def __len__(self):
        """
        Length of the ListMaster equal the number of childs appended.

        Returns:
            int: length of the ListMaster
        """
        return len(self.child_ids)

    def run_if_refresh(self):
        """
        Internal helper function the run if refresh function is called when the job status is 'refresh'. If the job was
        suspended previously, the job is going to be started again, to be continued.
        """
        self._logger.info(
            "{}, status: {}, finished: {} parallel master "
            "refresh".format(self.job_info_str, self.status, self.is_finished())
        )
        if self.is_finished():
            self.status.collect = True
            self.run()  # self.run_if_collect()
        elif (
            self.server.run_mode.non_modal or self.server.run_mode.queue
        ) and not self.submission_status.finished:
            self.run_static()
        else:
            self.refresh_job_status()
            if self.status.refresh:
                self.status.suspended = True
            if self.status.busy:
                self.status.refresh = True
                self.run_if_refresh()

    def _run_if_collect(self):
        """
        Internal helper function the run if collect function is called when the job status is 'collect'. It collects
        the simulation output using the standardized functions collect_output() and collect_logfiles(). Afterwards the
        status is set to 'finished'.
        """
        self._logger.info(
            "{}, status: {}, finished".format(self.job_info_str, self.status)
        )
        self.collect_output()

        job_id = self.get_job_id()
        db_dict = {}
        start_time = self.project.db.get_item_by_id(job_id)["timestart"]
        db_dict["timestop"] = datetime.now()
        db_dict["totalcputime"] = (db_dict["timestop"] - start_time).seconds
        self.project.db.item_update(db_dict, job_id)
        if not self.convergence_check():
            self.status.not_converged = True
        else:
            self.status.finished = True
        self._hdf5["status"] = self.status.string
        self._logger.info(
            "{}, status: {}, parallel master".format(self.job_info_str, self.status)
        )
        self.update_master()
        # self.send_to_database()

    def _run_if_new(self, debug=False):
        """
        Internal helper function the run if new function is called when the job status is 'initialized'. It prepares
        the hdf5 file and the corresponding directory structure.

        Args:
            debug (bool): Debug Mode
        """
        self.submission_status.submitted_jobs = 0
        super()._run_if_new(debug=debug)

    def convergence_check(self) -> bool:
        """
        Check if and all child jobs of the calculation are converged. May need be extended in the base classes depending
        on the specific application

        Returns:
             (bool): If the calculation is converged
        """
        for job in self.iter_jobs(convert_to_object=False):
            if job.status not in ["finished", "warning"]:
                return False
        return True

    def _validate_cores(self, job, cores_for_session):
        """
        Check if enough cores are available to start the next child job.

        Args:
            job (GenericJob): child job to be started
            cores_for_session (list): list of currently active cores - list of integers

        Returns:
            bool: [True/False]
        """
        return (
            self.get_child_cores() + job.server.cores + sum(cores_for_session)
            > self.server.cores
        )

    def _next_job_series(self, job):
        """
        Generate a list of child jobs to be executed in the next iteration.

        Args:
            job (GenericJob): child job to be started

        Returns:
            list: list of GenericJob objects
        """
        job_to_be_run_lst, cores_for_session = [], []
        while job is not None:
            self._logger.debug("create job: %s %s", job.job_info_str, job.master_id)
            if not job.status.finished:
                self.submission_status.submit_next()
                job_to_be_run_lst.append(job)
                cores_for_session.append(job.server.cores)
                self._logger.info(
                    "{}: finished job {}".format(self.job_name, job.job_name)
                )
            job = next(self._job_generator, None)
            if job is not None and self._validate_cores(job, cores_for_session):
                job = None
        return job_to_be_run_lst

    def _run_if_child_queue(self, job):
        """
        run function which is executed when the child jobs are submitted to the queue. In this case all child jobs are
        submitted at the same time without considering the number of cores specified for the Parallelmaster.

        Args:
            job (GenericJob): child job to be started
        """
        while job is not None:
            self._logger.debug("create job: %s %s", job.job_info_str, job.master_id)
            if not job.status.finished:
                job.run()
                self._logger.info(
                    "{}: submitted job {}".format(self.job_name, job.job_name)
                )
            job = next(self._job_generator, None)
        self.submission_status.submitted_jobs = self.submission_status.total_jobs
        self.status.suspended = True
        if self.is_finished():
            self.status.collect = True
            self.run()

    def _run_if_master_non_modal_child_non_modal(self, job):
        """
        run function which is executed when the Parallelmaster as well as its childs are running in non modal mode.

        Args:
            job (GenericJob): child job to be started
        """
        job_to_be_run_lst = self._next_job_series(job)
        if self.project.db.get_job_status(job_id=self.job_id) != "busy":
            self.status.suspended = True
            for job in job_to_be_run_lst:
                job.run()
            if self.master_id:
                del self
        else:
            self.run_static()

    def _run_if_master_modal_child_modal(self, job):
        """
        run function which is executed when the Parallelmaster as well as its childs are running in modal mode.

        Args:
            job (GenericJob): child job to be started
        """
        while job is not None:
            self._logger.debug("create job: %s %s", job.job_info_str, job.master_id)
            if not job.status.finished:
                self.submission_status.submit_next()
                job.run()
                self._logger.info(
                    "{}: finished job {}".format(self.job_name, job.job_name)
                )
            job = next(self._job_generator, None)
        if self.is_finished():
            self.status.collect = True
            self.run()
        elif self.status.busy:
            self.status.refresh = True
            self.run_if_refresh()
        else:
            self.status.suspended = True

    def _run_if_master_modal_child_non_modal(self, job):
        """
        run function which is executed when the Parallelmaster is running in modal mode and its childs are running in
        non modal mode.

        Args:
            job (GenericJob): child job to be started
        """
        pool = multiprocessing.Pool(self.server.cores)
        job_lst = []
        for i, p in enumerate(self._job_generator.parameter_list):
            if hasattr(self._job_generator, "job_name"):
                job = self.create_child_job(self._job_generator.job_name(parameter=p))
            else:
                job = self.create_child_job(self.ref_job.job_name + "_" + str(i))
            job = self._job_generator.modify_job(job=job, parameter=p)
            job.server.run_mode.modal = True
            job.save()
            job.project_hdf5.create_working_directory()
            job.write_input()
            if state.database.database_is_disabled or (
                state.queue_adapter is not None and state.queue_adapter.remote_flag
            ):
                job_lst.append(
                    (
                        job.project.path,
                        None,
                        job.project_hdf5.file_name + job.project_hdf5.h5_path,
                        False,
                        False,
                    )
                )
            else:
                job_lst.append((job.project.path, job.job_id, None, False, False))
        pool.starmap(job_wrapper_function, job_lst)
        if state.database.database_is_disabled:
            self.project.db.update()
        self.status.collect = True
        self.run()  # self.run_if_collect()

    def run_static(self):
        """
        The run_static function is executed within the GenericJob class and depending on the run_mode of the
        Parallelmaster and its child jobs a more specific run function is selected.
        """
        self._logger.info("{} run parallel master (modal)".format(self.job_info_str))
        self.status.running = True
        self.submission_status.total_jobs = len(self._job_generator)
        self.submission_status.submitted_jobs = 0
        if self.job_id and not self.is_finished():
            self._logger.debug(
                "{} child project {}".format(self.job_name, self.project.__str__())
            )
            job = next(self._job_generator, None)
            if job is not None:
                if (
                    self.server.run_mode.non_modal
                    or self.server.run_mode.queue
                    or self.server.run_mode.modal
                ) and job.server.run_mode.interactive:
                    self.run_if_interactive()
                elif self.server.run_mode.queue:
                    self._run_if_master_modal_child_non_modal(job=job)
                elif job.server.run_mode.queue:
                    self._run_if_child_queue(job)
                elif self.server.run_mode.non_modal and job.server.run_mode.non_modal:
                    self._run_if_master_non_modal_child_non_modal(job)
                elif (self.server.run_mode.modal and job.server.run_mode.modal) or (
                    self.server.run_mode.interactive and job.server.run_mode.interactive
                ):
                    self._run_if_master_modal_child_modal(job)
                elif self.server.run_mode.modal and job.server.run_mode.non_modal:
                    self._run_if_master_modal_child_non_modal(job)
                else:
                    raise TypeError()
        else:
            self.status.collect = True
            self.run()

    def run_if_interactive(self):
        if not (
            self.ref_job.server.run_mode.interactive
            or self.ref_job.server.run_mode.interactive_non_modal
        ):
            raise ValueError(
                "The child job has to be run_mode interactive or interactive_non_modal."
            )
        if isinstance(self.ref_job, GenericMaster):
            self.run_static()
        elif self.server.cores == 1:
            self.interactive_ref_job_initialize()
            for parameter in self._job_generator.parameter_list:
                self._job_generator.modify_job(job=self.ref_job, parameter=parameter)
                self.ref_job.run()
            self.ref_job.interactive_close()
        else:
            if self.server.cores > len(self._job_generator.parameter_list):
                number_of_jobs = len(self._job_generator.parameter_list)
            else:
                number_of_jobs = self.server.cores
            max_tasks_per_job = (
                int(len(self._job_generator.parameter_list) // number_of_jobs) + 1
            )
            parameters_sub_lst = [
                self._job_generator.parameter_list[i : i + max_tasks_per_job]
                for i in range(
                    0, len(self._job_generator.parameter_list), max_tasks_per_job
                )
            ]
            list_of_sub_jobs = [
                self.create_child_job("job_" + str(i)) for i in range(number_of_jobs)
            ]
            primary_job = list_of_sub_jobs[0]
            if not primary_job.server.run_mode.interactive_non_modal:
                raise ValueError(
                    "The child job has to be run_mode interactive_non_modal."
                )
            if primary_job.server.cores != 1:
                raise ValueError("The child job can only use a single core.")
            for iteration in range(len(parameters_sub_lst[0])):
                for job_ind, job in enumerate(list_of_sub_jobs):
                    if iteration < len(parameters_sub_lst[job_ind]):
                        self._job_generator.modify_job(
                            job=job, parameter=parameters_sub_lst[job_ind][iteration]
                        )
                        job.run()
                for job_ind, job in enumerate(list_of_sub_jobs):
                    if iteration < len(parameters_sub_lst[job_ind]):
                        job.interactive_fetch()
            for job in list_of_sub_jobs:
                job.interactive_close()
            self.interactive_ref_job_initialize()
            self.ref_job.run()
            for key in primary_job.interactive_cache.keys():
                output_sum = []
                for job in list_of_sub_jobs:
                    output = job["output/interactive/" + key]
                    if isinstance(output, np.ndarray):
                        output = output.tolist()
                    if isinstance(output, list):
                        output_sum += output
                    else:
                        raise TypeError(
                            "output should be list or numpy.ndarray but it is ",
                            type(output),
                        )
                self.ref_job.interactive_cache[key] = output_sum
            interactive_cache_backup = self.ref_job.interactive_cache.copy()
            self.ref_job.interactive_flush(path="generic", include_last_step=True)
            self.ref_job.interactive_cache = interactive_cache_backup
            self.ref_job.interactive_close()
        self.status.collect = True
        self.run()

    def create_child_job(self, job_name):
        """
        Internal helper function to create the next child job from the reference job template - usually this is called
        as part of the create_jobs() function.

        Args:
            job_name (str): name of the next job

        Returns:
            GenericJob: next job
        """
        project = self.child_project
        if not self.server.new_hdf:
            where_dict = {
                "job": str(job_name),
                "project": str(self.project_hdf5.project_path),
                "subjob": str(self.project_hdf5.h5_path + "/" + job_name),
            }
            response = self.project.db.get_items_dict(
                where_dict, return_all_columns=False
            )
            if len(response) > 0:
                job_id = response[-1]["id"]
            else:
                job_id = None
        else:
            job_id = project.get_job_id(job_specifier=job_name)
        if job_id is not None:
            ham = project.load(job_id)
            self._logger.debug("job {} found, status: {}".format(job_name, ham.status))
            if ham.server.run_mode.queue:
                self.project.refresh_job_status_based_on_job_id(job_id, que_mode=True)
            else:
                self.project.refresh_job_status_based_on_job_id(job_id, que_mode=False)
            if ham.status.aborted:
                ham.status.created = True

            self._logger.debug("job - status: {}".format(ham.status))
            return ham

        job = self.ref_job.copy()
        job = self._load_all_child_jobs(job_to_load=job)
        job.project_hdf5 = self.child_hdf(job_name)
        if isinstance(job, GenericMaster):
            for sub_job in job._job_object_dict.values():
                self._child_job_update_hdf(parent_job=job, child_job=sub_job)
        self._logger.debug(
            "create_job:: {} {} {} {}".format(
                self.project_hdf5.path,
                self._name,
                self.project_hdf5.h5_path,
                str(self.get_job_id()),
            )
        )
        job._name = job_name
        job.master_id = self.get_job_id()
        job.status.initialized = True
        if self.server.run_mode.non_modal and job.server.run_mode.modal:
            job.server.run_mode.non_modal = True
        elif self.server.run_mode.queue:
            job.server.run_mode.thread = True
        self._logger.info("{}: run job {}".format(self.job_name, job.job_name))
        return job

    def _db_server_entry(self):
        """
        connect all the info regarding the server into a single word that can be used e.g. as entry in a database

        Returns:
            (str): server info as single word

        """
        db_entry = super(ParallelMaster, self)._db_server_entry()
        if self.submission_status.total_jobs:
            return (
                db_entry
                + "#"
                + str(self.submission_status.submitted_jobs)
                + "/"
                + str(self.submission_status.total_jobs)
            )
        else:
            return db_entry + "#" + str(self.submission_status.submitted_jobs)

    def _run_if_repair(self):
        """
        Internal helper function the run if repair function is called when the run() function is called with the
        'repair' parameter.
        """
        reload_self = self.to_object()
        reload_self._run_if_created()

    def _init_child_job(self, parent):
        """
        Update our reference job and copy their run mode.

        Args:
            parent (:class:`.GenericJob`): job instance that this job was created from
        """
        super()._init_child_job(parent)
        if parent.server.run_mode.non_modal:
            self.server.run_mode.non_modal = True
        elif (
            parent.server.run_mode.interactive
            or parent.server.run_mode.interactive_non_modal
        ):
            self.server.run_mode.interactive = True


class GenericOutput(OrderedDict):
    """
    Generic Output just a place holder to store the output of the last child directly in the ParallelMaster.
    """

    def __init__(self):
        super(GenericOutput, self).__init__()


class JobGenerator(object):
    """
    Implements the functions to generate the parameter list, modify the individual jobs according to the parameter list
    and generate the new job names according to the parameter list.

    Subclasses have to override :meth:`.parameter_list()` to provide a list of (arbitrary) parameter objects and
    :meth:`.modify_job()` and may override :meth:`.job_name()` to provide custom job names.

    The generated jobs are created as child job from the given master.
    """

    def __init__(self, master):
        """
        Args:
            master (:class:`.ParallelMaster`): master job from which child jobs are created with
            :meth:`.ParallelMaster.create_child_job()`.
        """
        self._master = master
        self._childcounter = 0
        self._parameter_lst_cached = []

    @property
    def master(self):
        """
        :class:`.ParallelMaster`: the parallel master job with which this generator was initialized
        """
        return self._master

    @property
    @deprecate("use self.master instead")
    def _job(self):
        return self.master

    @property
    def parameter_list_cached(self):
        if len(self._parameter_lst_cached) == 0:
            self._parameter_lst_cached = self.parameter_list
        return self._parameter_lst_cached

    @property
    def parameter_list(self):
        """
        list:
            parameter objects passed to :meth:`.modify_job()` when the next
            job is requested.
        """
        raise NotImplementedError("Implement in derived class")

    @staticmethod
    def modify_job(job, parameter):
        """
        Modify next job with the parameter object.  job is already the newly
        created job object cloned from the template job, so this function has
        to return the same instance, but may (and should) modify it.

        Args:
            job (:class:`.GenericJob`):
                new job instance
            parameter (type):
                current parameter object drawn from :attr:`.parameter_list`.

        Returns:
            :class:`.GenericJob`: must be the given job
        """
        raise NotImplementedError("Implement in derived class")

    def job_name(self, parameter):
        """
        Return new job name from parameter object.  The next child job created
        will have this name.  Subclasses may override this to give custom job
        names.

        Args:
            parameter (type):
                current parameter object drawn from :attr:`.parameter_list`.

        Returns:
            str: job name for the next child job
        """
        return self._master.ref_job.job_name + "_" + str(self._childcounter)

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def __len__(self):
        return len(self.parameter_list_cached)

    def next(self):
        """
        Iterate over the child jobs

        Returns:
            :class:`~.GenericJob`: new job object
        """
        if len(self.parameter_list_cached) > self._childcounter:
            current_paramenter = self.parameter_list_cached[self._childcounter]
            job = self._master.create_child_job(
                self.job_name(parameter=current_paramenter)
            )
            if job is not None:
                self._childcounter += 1
                job = self.modify_job(job=job, parameter=current_paramenter)
                return job
            else:
                raise StopIteration()
        else:
            self._master.refresh_job_status()
            raise StopIteration()



"""
Efficient storage of ragged arrays in a flattened format.
"""

# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

__author__ = ("Marvin Poul", "Niklas Leimeroth")
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "production"
__date__ = "Jul 16, 2020"


import copy
import warnings
from typing import Callable, Iterable, List, Tuple, Any

import numpy as np
import h5py
from pyiron_base.interfaces.has_hdf import HasHDF


def _ensure_str_array_size(array, strlen):
    """
    Ensures that the given array can store at least string of length `strlen`.

    Args:
        array (ndarray): array of dtype <U
        strlen (int): maximum length that should fit in it
    Returns:
        ndarray: either `array` or resized copy
    """
    current_length = array.itemsize // np.dtype("1U").itemsize
    if current_length < strlen:
        return array.astype(f"{2 * strlen}U")
    else:
        return array


class FlattenedStorage(HasHDF):
    """
    Efficient storage of ragged arrays in flattened arrays.

    This class stores multiple arrays at the same time.  Storage is organized in "chunks" that may be of any size, but
    all arrays within chunk are of the same size, e.g.

    >>> a = [ [1], [2, 3], [4,  5,  6] ]
    >>> b = [ [2], [4, 6], [8, 10, 12] ]

    are stored as in three chunks like

    >>> a_flat = [ 1,  2, 3,  4,  5,  6 ]
    >>> b_flat = [ 2,  4, 6,  8, 10, 12 ]

    with additional metadata to indicate where the boundaries of each chunk are.

    First add arrays and chunks like this

    >>> store = FlattenedStorage()
    >>> store.add_array("even", dtype=np.int64)
    >>> store.add_chunk(1, even=[2])
    >>> store.add_chunk(2, even=[4,  6])
    >>> store.add_chunk(3, even=[8, 10, 12])

    where the first argument indicates the length of each chunk.  You may retrieve stored values like this

    >>> store.get_array("even", 1)
    array([4, 6])
    >>> store.get_array("even", 0)
    array([2])

    where the second arguments are integer indices in the order of insertion.  After intial storage you may modify
    arrays.

    >>> store.set_array("even", 0, [0])
    >>> store.get_array("even", 0)
    array([0])

    As a shorthand you can use regular index syntax

    >>> store["even", 0] = [2]
    >>> store["even", 0]
    array([2])
    >>> store["even", 1]
    array([4, 6])
    >>> store["even"]
    array([2, 4, 6, 8, 10, 12])
    >>> store["even", 0] = [0]

    You can add arrays to the storage even after you added already other arrays and chunks.

    >>> store.add_array("odd", dtype=np.int64, fill=0)
    >>> store.get_array("odd", 1)
    array([0, 0])
    >>> store.set_array("odd", 0, [1])
    >>> store.set_array("odd", 1, [3, 5])
    >>> store.set_array("odd", 2, [7, 9, 11])
    >>> store.get_array("odd", 2)
    array([ 7,  9, 11])

    Because the second chunk is already known to be of length two and `fill` was specified the 'odd' array has been
    appropriatly allocated.

    Additionally arrays may also only have one value per chunk ("per chunk", previous examples are "per element").

    >>> store.add_array("sum", dtype=np.int64, per="chunk")
    >>> for i in range(len(store)):
    ...    store.set_array("sum", i, sum(store.get_array("even", i) + store.get_array("odd", i)))
    >>> store.get_array("sum", 0)
    1
    >>> store.get_array("sum", 1)
    18
    >>> store.get_array("sum", 2)
    57

    Finally you may add multiple arrays in one call to :method:`.add_chunk` by using keyword arguments

    >>> store.add_chunk(4, even=[14, 16, 18, 20], odd=[13, 15, 17, 19], sum=119)
    >>> store.get_array("sum", 3)
    119
    >>> store.get_array("even", 3)
    array([14, 16, 18, 20])

    It is usually not necessary to call :method:`.add_array` before :method:`.add_chunk`, the type of the array will be
    inferred in this case.

    If you skip the `frame` argument to :meth:`.get_array` it will return a flat array of all the values for that array
    in storage.

    >>> store.get_array("sum")
    array([  1,  18,  57, 119])
    >>> store.get_array("even")
    array([ 0,  4,  6,  8, 10, 12, 14, 16, 18, 20])

    Arrays may be of more complicated shape, too, see :method:`.add_array` for details.

    Use :method:`.copy` to obtain a deep copy of the storage, for shallow copies using the builting `copy.copy` is
    sufficient.

    >>> copy = store.copy()
    >>> copy["even", 0]
    array([0])
    >>> copy["even", 1]
    array([4, 6])
    >>> copy["even"]
    array([0, 4, 6, 8, 10, 12])

    Storages can be :method:`.split` and :method:`.join` again as long as their internal chunk structure is consistent,
    i.e. same number of chunks and same chunk lengths.  If this is not the case a `ValueError` is raised.

    >>> even = store.split(["even"])
    >>> bool(even.has_array("even"))
    True
    >>> bool(even.has_array("odd"))
    False
    >>> odd = store.split(["odd"])

    :method:`.join` adds new arrays to the storage it is called on in-place.  To leave it unchanged, simply call copy
    before join.
    >>> both = even.copy().join(odd)

    Chunks may be given string names, either by passing `identifier` to :method:`.add_chunk` or by setting to the
    special per chunk array "identifier"

    >>> store.set_array("identifier", 1, "second")
    >>> all(store.get_array("even", "second") == store.get_array("even", 1))
    True

    When adding new arrays follow the convention that per-structure arrays should be named in singular and per-atom
    arrays should be named in plural.

    You may initialize flattened storage objects with a ragged lists or numpy arrays of dtype object

    >>> even = [ list(range(0, 2, 2)), list(range(2, 6, 2)), list(range(6, 12, 2)) ]
    >>> even
    [[0], [2, 4], [6, 8, 10]]

    >>> import numpy as np
    >>> odd = np.array([ np.arange(1, 2, 2), np.arange(3, 6, 2), np.arange(7, 12, 2) ], dtype=object)
    >>> odd
    array([array([1]), array([3, 5]), array([ 7,  9, 11])], dtype=object)

    >>> store = FlattenedStorage(even=even, odd=odd)
    >>> store.get_array("even", 1)
    array([2, 4])
    >>> store.get_array("odd", 2)
    array([ 7,  9, 11])
    >>> len(store)
    3
    """

    __version__ = "0.2.0"
    __hdf_version__ = "0.3.0"
    _default_fill_values = {
        np.dtype("int8"): -1,
        np.dtype("int16"): -1,
        np.dtype("int32"): -1,
        np.dtype("int64"): -1,
        np.dtype("float16"): np.nan,
        np.dtype("float32"): np.nan,
        np.dtype("float64"): np.nan,
        np.dtype("object"): None,
        np.dtype("uint8"): 0,
        np.dtype("uint16"): 0,
        np.dtype("uint32"): 0,
        np.dtype("uint64"): 0,
        str: "_default",
    }

    def __init__(self, num_chunks=1, num_elements=1, **kwargs):
        """
        Create new flattened storage.

        Args:
            num_chunks (int): pre-allocation for per chunk arrays
            num_elements (int): pre-allocation for per elements arrays
        """
        # tracks allocated versed as yet used number of chunks/elements
        self._num_chunks_alloc = num_chunks
        self._num_elements_alloc = num_elements
        self.num_chunks = 0
        self.num_elements = 0
        # store the starting index for properties with unknown length
        self.current_element_index = 0
        # store the index for properties of known size, stored at the same index as the chunk
        self.current_chunk_index = 0
        # Also store indices of chunk recently added
        self.prev_chunk_index = 0
        self.prev_element_index = 0
        self._fill_values = {}

        self._init_arrays()

        if len(kwargs) == 0:
            return

        if len(set(len(chunks) for chunks in kwargs.values())) != 1:
            raise ValueError("Not all initializers provide the same number of chunks!")
        keys = kwargs.keys()
        for chunk_list in zip(*kwargs.values()):
            chunk_length = len(chunk_list[0])
            # values in chunk_list may either be a sequence of chunk_length, scalars (see hasattr check) or a sequence of
            # length 1
            if any(
                hasattr(c, "__len__") and len(c) != chunk_length and len(c) != 1
                for c in chunk_list
            ):
                raise ValueError("Inconsistent chunk length in initializer!")
            self.add_chunk(chunk_length, **{k: c for k, c in zip(keys, chunk_list)})

    def _init_arrays(self):
        self._per_element_arrays = {}

        self._per_chunk_arrays = {
            "start_index": np.full(
                self._num_chunks_alloc, dtype=np.int32, fill_value=0
            ),
            "length": np.full(self._num_chunks_alloc, dtype=np.int32, fill_value=0),
            "identifier": np.empty(self._num_chunks_alloc, dtype=np.dtype("U20")),
        }

    def __len__(self):
        return self.current_chunk_index

    def copy(self):
        """
        Return a deep copy of the storage.

        Returns:
            :class:`.FlattenedStorage`: copy of self
        """
        return copy.deepcopy(self)

    def find_chunk(self, identifier):
        """
        Return integer index for given identifier.

        Args:
            identifier (str): name of chunk previously passed to :method:`.add_chunk`

        Returns:
            int: integer index for chunk

        Raises:
            KeyError: if identifier is not found in storage
        """
        for i, name in enumerate(self._per_chunk_arrays["identifier"]):
            if name == identifier:
                return i
        raise KeyError(f"No chunk named {identifier}")

    def _get_per_element_slice(self, frame):
        start = self._per_chunk_arrays["start_index"][frame]
        end = start + self._per_chunk_arrays["length"][frame]
        return slice(start, end, 1)

    def _resize_elements(self, new):
        old_max = self._num_elements_alloc
        self._num_elements_alloc = new
        for k, a in self._per_element_arrays.items():
            new_shape = (new,) + a.shape[1:]
            try:
                a.resize(new_shape)
            except ValueError:
                self._per_element_arrays[k] = np.resize(a, new_shape)
        if old_max < new:
            for k in self._per_element_arrays.keys():
                if k in self._fill_values.keys():
                    self._per_element_arrays[k][old_max:] = self._fill_values[k]

    def _resize_chunks(self, new):
        old_max = self._num_chunks_alloc
        self._num_chunks_alloc = new
        for k, a in self._per_chunk_arrays.items():
            new_shape = (new,) + a.shape[1:]
            try:
                a.resize(new_shape)
            except ValueError:
                self._per_chunk_arrays[k] = np.resize(a, new_shape)
        if old_max < new:
            for k in self._per_chunk_arrays.keys():
                if k in self._fill_values.keys():
                    self._per_chunk_arrays[k][old_max:] = self._fill_values[k]

    def add_array(self, name, shape=(), dtype=np.float64, fill=None, per="element"):
        """
        Add a custom array to the container.

        When adding an array after some chunks have been added, specifying `fill` will be used as a default value
        for the value of the array for those chunks.

        Adding an array with the same name twice is ignored, if dtype and shape match, otherwise raises an exception.

        >>> store = FlattenedStorage()
        >>> store.add_chunk(1, "foo")
        >>> store.add_array("energy", shape=(), dtype=np.float64, fill=42, per="chunk")
        >>> store.get_array("energy", 0)
        42.0

        Args:
            name (str): name of the new array
            shape (tuple of int): shape of the new array per element or chunk; scalars can pass ()
            dtype (type): data type of the new array, string arrays can pass 'U$n' where $n is the length of the string
            fill (object): populate the new array with this value for existing chunk, if given; default `None`
            per (str): either "element" or "chunk"; denotes whether the new array should exist for every element in a
                       chunk or only once for every chunk; case-insensitive

        Raises:
            ValueError: if wrong value for `per` is given
            ValueError: if array with same name but different parameters exists already
        """

        if per == "structure":
            per = "chunk"
            warnings.warn(
                'per="structure" is deprecated, use pr="chunk"',
                category=DeprecationWarning,
                stacklevel=2,
            )
        if per == "atom":
            per = "element"
            warnings.warn(
                'per="atom" is deprecated, use pr="element"',
                category=DeprecationWarning,
                stacklevel=2,
            )

        if name in self._per_element_arrays:
            a = self._per_element_arrays[name]
            if (
                a.shape[1:] != shape
                or not np.can_cast(dtype, a.dtype)
                or per != "element"
            ):
                raise ValueError(
                    f"Array with name '{name}' exists with shape {a.shape[1:]} and dtype {a.dtype}."
                )
            else:
                return

        if name in self._per_chunk_arrays:
            a = self._per_chunk_arrays[name]
            if (
                a.shape[1:] != shape
                or not np.can_cast(dtype, a.dtype)
                or per != "chunk"
            ):
                raise ValueError(
                    f"Array with name '{name}' exists with shape {a.shape[1:]} and dtype {a.dtype}."
                )
            else:
                return

        per = per.lower()
        if per == "element":
            shape = (self._num_elements_alloc,) + shape
            store = self._per_element_arrays
        elif per == "chunk":
            shape = (self._num_chunks_alloc,) + shape
            store = self._per_chunk_arrays
        else:
            raise ValueError(f'per must "element" or "chunk", not {per}')

        if fill is None:
            store[name] = np.empty(shape=shape, dtype=dtype)
        else:
            store[name] = np.full(shape=shape, fill_value=fill, dtype=dtype)

        if fill is None and store[name].dtype in self._default_fill_values:
            fill = self._default_fill_values[store[name].dtype]
        if fill is not None:
            self._fill_values[name] = fill

    def get_array(self, name, frame=None):
        """
        Fetch array for given structure.

        Works for per atom and per arrays.

        Args:
            name (str): name of the array to fetch
            frame (int, str, optional): selects structure to fetch, as in :method:`.get_structure()`, if not given
                                        return a flat array of all values for either all chunks or elements

        Returns:
            :class:`numpy.ndarray`: requested array

        Raises:
            `KeyError`: if array with name does not exists
        """

        if isinstance(frame, str):
            frame = self.find_chunk(frame)
        if name in self._per_element_arrays:
            if frame is not None:
                return self._per_element_arrays[name][
                    self._get_per_element_slice(frame)
                ]
            else:
                return self._per_element_arrays[name][: self.num_elements]
        elif name in self._per_chunk_arrays:
            if frame is not None:
                return self._per_chunk_arrays[name][frame]
            else:
                return self._per_chunk_arrays[name][: self.num_chunks]
        else:
            raise KeyError(f"no array named {name}")

    def get_array_ragged(self, name: str) -> np.ndarray:
        """
        Return elements of array `name` in all chunks.  Values are returned in a ragged array of dtype=object.

        If `name` specifies a per chunk array, there's nothing to pad and this method is equivalent to
        :method:`.get_array`.

        Args:
            name (str): name of array to fetch

        Returns:
            numpy.ndarray, dtype=object: ragged arrray of all elements in all chunks
        """
        if name in self._per_chunk_arrays:
            return self.get_array(name)
        # pre-allocated as dtype=object, then setting individual elements makes sure that element arrays retain their
        # dtype
        result = np.empty(len(self), dtype=object)
        for i in range(len(self)):
            result[i] = self.get_array(name, i)
        return result

    def get_array_filled(self, name: str) -> np.ndarray:
        """
        Return elements of array `name` in all chunks.  Arrays are padded to be all of the same length.

        The padding value depends on the datatpye of the array or can be configured via the `fill` parameter of
        :method:`.add_array`.

        If `name` specifies a per chunk array, there's nothing to pad and this method is equivalent to
        :method:`.get_array`.

        Args:
            name (str): name of array to fetch

        Returns:
            numpy.ndarray: padded arrray of all elements in all chunks
        """
        if name in self._per_chunk_arrays:
            return self.get_array(name)
        values = self.get_array_ragged(name)
        max_len = self._per_chunk_arrays["length"].max()

        def resize_and_pad(v):
            l = len(v)
            per_shape = self._per_element_arrays[name].shape[1:]
            v = np.resize(v, max_len * np.prod(per_shape, dtype=int))
            v = v.reshape((max_len,) + per_shape)
            if name in self._fill_values:
                fill = self._fill_values[name]
            else:
                fill = np.zeros(1, dtype=self._per_element_arrays[name].dtype)[0]
            v[l:] = fill
            return v

        return np.array([resize_and_pad(v) for v in values])

    def set_array(self, name, frame, value):
        """
        Add array for given structure.

        Works for per chunk and per element arrays.

        Args:
            name (str): name of array to set
            frame (int, str): selects structure to set, as in :method:`.get_strucure()`
            value: value (for per chunk) or array of values (for per element); type and shape as per :meth:`.hasarray()`.

        Raises:
            `KeyError`: if array with name does not exists
        """

        if isinstance(frame, str):
            frame = self.find_chunk(frame)
        if name in self._per_element_arrays:
            if self._per_element_arrays[name].dtype.char == "U":
                self._per_element_arrays[name] = _ensure_str_array_size(
                    self._per_element_arrays[name], max(map(len, value))
                )
            self._per_element_arrays[name][self._get_per_element_slice(frame)] = value
        elif name in self._per_chunk_arrays:
            if self._per_chunk_arrays[name].dtype.char == "U":
                if isinstance(value, np.ndarray) and value.ndim == 0:
                    strlen = len(value.item())
                else:
                    strlen = len(value)
                self._per_chunk_arrays[name] = _ensure_str_array_size(
                    self._per_chunk_arrays[name], strlen
                )
            self._per_chunk_arrays[name][frame] = value
        else:
            raise KeyError(f"no array named {name}")

    def __getitem__(self, index):
        if isinstance(index, tuple) and len(index) == 2:
            return self.get_array(index[0], index[1])
        else:
            return self.get_array(index)

    def __setitem__(self, index, value):
        if isinstance(index, tuple) and len(index) == 2:
            self.set_array(index[0], index[1], value)
        else:
            raise IndexError("Must specify chunk index.")

    def has_array(self, name):
        """
        Checks whether an array of the given name exists and returns meta data given to :method:`.add_array()`.

        >>> container.has_array("energy")
        {'shape': (), 'dtype': np.float64, 'per': 'chunk'}
        >>> container.has_array("fnorble")
        None

        Args:
            name (str): name of the array to check

        Returns:
            None: if array does not exist
            dict: if array exists, keys corresponds to the shape, dtype and per arguments of :method:`.add_array`
        """
        if name in self._per_element_arrays:
            a = self._per_element_arrays[name]
            per = "element"
        elif name in self._per_chunk_arrays:
            a = self._per_chunk_arrays[name]
            per = "chunk"
        else:
            return None
        return {"shape": a.shape[1:], "dtype": a.dtype, "per": per}

    def list_arrays(self) -> List[str]:
        """
        Return a list of names of arrays inside the storage.

        Returns:
            list of str: array names
        """
        return list(self._per_chunk_arrays) + list(self._per_element_arrays)

    def sample(
        self, selector: Callable[["FlattenedStorage", int], bool]
    ) -> "FlattenedStorage":
        """
        Create a new storage with chunks selected by given function.

        If called on a subclass this correctly returns an instance of that subclass instead.

        Args:
            select (callable): function that takes this storage as the first argument and the chunk index to sample as
                               the second argument; if it returns True it will be part of the new storage.

        Returns:
            :class:`.FlattenedStorage` or subclass: storage with the selected chunks
        """
        new = self.__class__()
        for k, a in self._per_chunk_arrays.items():
            if k not in ("start_index", "length", "identifier"):
                new.add_array(k, shape=a.shape[1:], dtype=a.dtype, per="chunk")
        for k, a in self._per_element_arrays.items():
            new.add_array(k, shape=a.shape[1:], dtype=a.dtype, per="element")
        for i in range(len(self)):
            if selector(self, i):
                new.add_chunk(
                    self.get_array("length", i),
                    identifier=self.get_array("identifier", i),
                )
                for k in self._per_chunk_arrays:
                    if k not in ("start_index", "length", "identifier"):
                        new.set_array(k, len(new) - 1, self.get_array(k, i))
                for k in self._per_element_arrays:
                    new.set_array(k, len(new) - 1, self.get_array(k, i))
        return new

    def split(self, array_names: Iterable[str]) -> "FlattenedStorage":
        """
        Return a new storage with only the selected arrays present.

        Arrays are deep-copied from `self`.

        Args:
            array_names (list of str): names of the arrays to present in new storage

        Returns:
            :class:`.FlattenedStorage`: storage with split arrays
        """
        for k in array_names:
            if k not in self._per_element_arrays and k not in self._per_chunk_arrays:
                raise ValueError(f"Array name {k} not present in FlattenedStorage!")

        split = copy.copy(self)
        for k in list(split._per_element_arrays):
            if k not in array_names:
                del split._per_element_arrays[k]
            else:
                split._per_element_arrays[k] = np.copy(split._per_element_arrays[k])
        for k in list(split._per_chunk_arrays):
            if k not in array_names and k not in (
                "start_index",
                "length",
                "identifier",
            ):
                del split._per_chunk_arrays[k]
            else:
                split._per_chunk_arrays[k] = np.copy(split._per_chunk_arrays[k])
        return split

    def join(
        self, store: "FlattenedStorage", lsuffix: str = "", rsuffix: str = ""
    ) -> "FlattenedStorage":
        """
        Merge given storage into this one.

        `self` and `store` may not share any arrays.  Arrays defined on `stores` are copied and then added to `self`.

        Args:
            store (:class:`.FlattenedStorage`): storage to join

        Returns:
            :class:`.FlattenedStorage`: self
        """
        if len(self) != len(store):
            raise ValueError(
                "FlattenedStorages to be joined have to be of the same length!"
            )
        if (self["length"] != store["length"]).any():
            raise ValueError(
                "FlattenedStorages to be joined have to have same length chunks everywhere!"
            )
        if lsuffix == rsuffix != "":
            raise ValueError("lsuffix and rsuffix may not be equal!")
        rename = lsuffix != "" or rsuffix != ""
        if not rename:
            shared_elements = set(self._per_element_arrays).intersection(
                store._per_element_arrays
            )
            shared_chunks = set(self._per_chunk_arrays).intersection(
                store._per_chunk_arrays
            )
            shared_chunks.remove("start_index")
            shared_chunks.remove("length")
            shared_chunks.remove("identifier")
            if len(shared_elements) > 0 or len(shared_chunks) > 0:
                raise ValueError(
                    "FlattenedStorages to be joined may have common arrays only if lsuffix or rsuffix are given!"
                )

        for k, a in store._per_element_arrays.items():
            if k in self._per_element_arrays and rename:
                self._per_element_arrays[k + lsuffix] = self._per_element_arrays[k]
                k += rsuffix
            self._per_element_arrays[k] = a

        for k, a in store._per_chunk_arrays.items():
            if k not in ("start_index", "length", "identifier"):
                if k in self._per_chunk_arrays and rename:
                    self._per_chunk_arrays[k + lsuffix] = self._per_chunk_arrays[k]
                    k += rsuffix
                self._per_chunk_arrays[k] = a

        self._resize_elements(self._num_elements_alloc)
        self._resize_chunks(self._num_chunks_alloc)
        return self

    def add_chunk(self, chunk_length, identifier=None, **arrays):
        """
        Add a new chunk to the storeage.

        Additional keyword arguments given specify arrays to store for the chunk.  If an array with the given keyword
        name does not exist yet, it will be added to the container.

        >>> container = FlattenedStorage()
        >>> container.add_chunk(2, identifier="A", energy=3.14)
        >>> container.get_array("energy", 0)
        3.14

        If the first axis of the extra array matches the length of the chunk, it will be added as an per element array,
        otherwise as an per chunk array.

        >>> container.add_chunk(2, identifier="B", forces=2 * [[0,0,0]])
        >>> len(container.get_array("forces", 1)) == 2
        True

        Reshaping the array to have the first axis be length 1 forces the array to be set as per chunk array.  That axis
        will then be stripped.

        >>> container.add_chunk(2, identifier="C", pressure=np.eye(3)[np.newaxis, :, :])
        >>> container.get_array("pressure", 2).shape
        (3, 3)

        .. attention: Edge-case!

            This will not work when the chunk length is also 1 and the array does not exist yet!  In this case the array
            will be assumed to be per element and there is no way around explicitly calling :method:`.add_array()`.


        Args:
            chunk_length (int): length of the new chunk
            identifier (str, optional): human-readable name for the chunk, if None use current chunk index as string
            **kwargs: additional arrays to store for the chunk
        """

        if identifier is None:
            identifier = str(self.num_chunks)

        n = chunk_length
        new_elements = self.current_element_index + n

        if new_elements > self._num_elements_alloc:
            self._resize_elements(max(new_elements, self._num_elements_alloc * 2))
        if self.current_chunk_index + 1 > self._num_chunks_alloc:
            self._resize_chunks(max(1, self._num_chunks_alloc * 2))

        if new_elements > self.num_elements:
            self.num_elements = new_elements
        if self.current_chunk_index + 1 > self.num_chunks:
            self.num_chunks += 1

        # len of chunk to index into the initialized arrays
        i = self.current_element_index + n

        self._per_chunk_arrays["start_index"][
            self.current_chunk_index
        ] = self.current_element_index
        self._per_chunk_arrays["length"][self.current_chunk_index] = n
        self._per_chunk_arrays["identifier"] = _ensure_str_array_size(
            self._per_chunk_arrays["identifier"], len(identifier)
        )
        self._per_chunk_arrays["identifier"][self.current_chunk_index] = identifier

        for k, a in arrays.items():
            a = np.asarray(a)
            if k not in self._per_element_arrays and k not in self._per_chunk_arrays:
                if len(a.shape) > 0 and a.shape[0] == n:
                    self.add_array(k, shape=a.shape[1:], dtype=a.dtype, per="element")
                else:
                    shape = a.shape
                    # if the first axis was added by the caller to force to add a per chunk array, remove it again here
                    if len(shape) > 0 and a.shape[0] == 1:
                        shape = shape[1:]
                    self.add_array(k, shape=shape, dtype=a.dtype, per="chunk")
            # same as above: if the first axis was added by the caller to force to add a per chunk array, remove it
            # again here
            if k in self._per_chunk_arrays and len(a.shape) > 0 and a.shape[0] == 1:
                a = a[0]
            self.set_array(k, self.current_chunk_index, a)

        self.prev_chunk_index = self.current_chunk_index
        self.prev_element_index = self.current_element_index

        # Set new current_element_index and increase current_chunk_index
        self.current_chunk_index += 1
        self.current_element_index = i
        # return last_chunk_index, last_element_index

    def extend(self, other: "FlattenedStorage"):
        self._check_compatible_fill_values(other=other)

        combined_num_chunks = self.num_chunks + other.num_chunks
        combined_num_elements = self.num_elements + other.num_elements
        if combined_num_chunks > self._num_chunks_alloc:
            self._resize_chunks(combined_num_chunks)
        if combined_num_elements > self._num_elements_alloc:
            self._resize_elements(combined_num_elements)

        for k, a in other._per_chunk_arrays.items():
            # add start_index of last chunk to start_index of other for correct mapping
            if (
                k == "start_index" and len(self) > 0
            ):  # Check if len > 0 to ensure that no random values are accessed for length and start_index after empty init
                last = self.num_chunks - 1
                len_last = self._per_chunk_arrays["length"][last]
                a = (
                    a + self._per_chunk_arrays[k][last] + len_last
                )  # no += to prevent inplace mutation
            if k not in self._per_chunk_arrays.keys():
                dtype, fill = get_dtype_and_fill(storage=other, name=k)
                self.add_array(
                    name=k, dtype=dtype, shape=a.shape[1:], fill=fill, per="chunk"
                )
            self._per_chunk_arrays[k][self.num_chunks : combined_num_chunks] = a[
                0 : other.num_chunks
            ]

        for k, a in other._per_element_arrays.items():
            if k not in self._per_element_arrays.keys():
                dtype, fill = get_dtype_and_fill(storage=other, name=k)
                self.add_array(
                    name=k, shape=a.shape[1:], dtype=dtype, fill=fill, per="element"
                )
            self._per_element_arrays[k][self.num_elements : combined_num_elements] = a[
                0 : other.num_elements
            ]
        self.num_elements = combined_num_elements
        self.num_chunks = combined_num_chunks
        self.current_chunk_index = self.num_chunks
        self.current_element_index = self.num_elements

    def _check_compatible_fill_values(self, other: "FlattenedStorage"):
        """
        Check if fill values of 2 FlattenedStorages match to prevent errors due to wrong fill values,
        f.e. after extending to the storage.

        Args:
            other (FlattenedStorage): Another FlattenedStorage instance

        Raises:
            ValueError: Raises when the storages have different fill values for a key
        """
        for k in set(self._fill_values).intersection(other._fill_values):
            if np.isnan(self._fill_values[k]) and np.isnan(other._fill_values[k]):
                continue
            else:
                if self._fill_values[k] != other._fill_values[k]:
                    raise ValueError(
                        "Fill values for arrays in storages don't match, can't perform requested operation"
                    )

    def _get_hdf_group_name(self):
        return "flat_storage"

    def _to_hdf(self, hdf):
        def write_array(name, array, hdf):
            if array.dtype.char == "U":
                # numpy stores unicode data in UTF-32/UCS-4, but h5py wants UTF-8, so we manually encode them here
                # TODO: string arrays with shape != () not handled
                hdf[name] = np.array(
                    [s.encode("utf8") for s in array],
                    # each character in a utf8 string might be encoded in up to 4 bytes, so to
                    # make sure we can store any string of length n we tell h5py that the
                    # string will be 4 * n bytes; numpy's dtype does this calculation already
                    # in itemsize, so we don't need to repeat it here
                    # see also https://docs.h5py.org/en/stable/strings.html
                    dtype=h5py.string_dtype("utf8", array.dtype.itemsize),
                )
            else:
                hdf[name] = array

        # truncate arrays to necessary size before writing
        self._resize_elements(self.num_elements)
        self._resize_chunks(self.num_chunks)

        hdf["num_elements"] = self._num_elements_alloc
        hdf["num_chunks"] = self._num_chunks_alloc

        hdf_arrays = hdf.open("element_arrays")
        for k, a in self._per_element_arrays.items():
            write_array(k, a, hdf_arrays)

        hdf_arrays = hdf.open("chunk_arrays")
        for k, a in self._per_chunk_arrays.items():
            write_array(k, a, hdf_arrays)

        hdf["_fill_values"] = self._fill_values

    def _from_hdf(self, hdf, version=None):
        def read_array(name, hdf):
            a = np.asarray(hdf[name])
            if a.dtype.char == "S":
                # if saved as bytes, we wrote this as an encoded unicode string, so manually decode here
                # TODO: string arrays with shape != () not handled
                a = np.fromiter(
                    (s.decode("utf8") for s in a),
                    # itemsize of original a is four bytes per character, so divide by four to get
                    # length of the orignal stored unicode string; np.dtype('U1').itemsize is just a
                    # platform agnostic way of knowing how wide a unicode charater is for numpy
                    dtype=f"U{a.dtype.itemsize//np.dtype('U1').itemsize}",
                )
            return a

        try:
            num_chunks = hdf["num_chunks"]
            num_elements = hdf["num_elements"]
        except ValueError:
            num_chunks = hdf["num_structures"]
            num_elements = hdf["num_atoms"]

        self._num_chunks_alloc = self.num_chunks = self.current_chunk_index = num_chunks
        self._num_elements_alloc = (
            self.num_elements
        ) = self.current_element_index = num_elements

        if version == "0.1.0":
            with hdf.open("arrays") as hdf_arrays:
                for k in hdf_arrays.list_nodes():
                    a = read_array(k, hdf_arrays)
                    if a.shape[0] == self._num_elements_alloc:
                        self._per_element_arrays[k] = a
                    elif a.shape[0] == self._num_chunks_alloc:
                        self._per_chunk_arrays[k] = a
        elif version == "0.2.0" or "0.3.0":
            with hdf.open("element_arrays") as hdf_arrays:
                for k in hdf_arrays.list_nodes():
                    self._per_element_arrays[k] = read_array(k, hdf_arrays)
            with hdf.open("chunk_arrays") as hdf_arrays:
                for k in hdf_arrays.list_nodes():
                    self._per_chunk_arrays[k] = read_array(k, hdf_arrays)
        else:
            raise RuntimeError(
                f"Unsupported HDF version {version}; use an older version of pyiron to load this job!"
            )

        for k, a in self._per_chunk_arrays.items():
            if a.shape[0] != self._num_chunks_alloc:
                raise RuntimeError(
                    f"per-chunk array {k} read inconsistently from HDF: "
                    f"shape {a.shape[0]} does not match global allocation {self._num_chunks_alloc}!"
                )
        for k, a in self._per_element_arrays.items():
            if a.shape[0] != self._num_elements_alloc:
                raise RuntimeError(
                    f"per-element array {k} read inconsistently from HDF: "
                    f"shape {a.shape[0]} does not match global allocation {self._num_elements_alloc}!"
                )

        if version >= "0.3.0":
            self._fill_values = hdf["_fill_values"]


def get_dtype_and_fill(storage: FlattenedStorage, name: str) -> Tuple[np.generic, Any]:
    fill = None
    if name in storage._fill_values.keys():
        fill = storage._fill_values[name]
        dtype = type(fill)
    else:
        a = storage.get_array(name)
        if a.dtype.char == "U":
            dtype = str
        else:
            dtype = a.dtype
        try:
            fill = FlattenedStorage._default_fill_values[dtype]
        except KeyError:
            raise ValueError(
                f"Could not determine a default fill value for array {name}"
            )
    return dtype, fill



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
The `state` module holds (almost!) all the code for defining the global state of a pyiron instance.
Such "global" behaviour is achieved by using the `Singleton` metaclass to guarantee that each class only even has a
single instance per session.
These are all instantiated for the first time inside their respective module for easy access, and collected here in the
init under the `state` object to give a single, even easier point of access.

Here's the "almost": Right now the database management still lives off in its own module but is referenced here and
ultimately should probably be relocated here (work is ongoing on our database interaction...), and there is a
`JobTypeChoice` class that is anyhow on the chopping block and will be deleted once we don't need it for backwards
compatibility.
"""

from pyiron_base.state.logger import logger as _logger
from pyiron_base.database.manager import database as _database
from pyiron_base.state.publications import publications as _publications
from pyiron_base.state.queue_adapter import queue_adapters as _queue_adapters
from pyiron_base.state.settings import settings as _settings
from pyiron_base.interfaces.singleton import Singleton
from typing import Dict, Union

__author__ = "Liam Huber"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "production"
__date__ = "Oct 22, 2021"


class State(metaclass=Singleton):
    """
    A helper class to give quick and easy access to all the singleton classes which together define the state module.

    Attributes:
        logger: Self-explanatory.
        publications: Bibliography of papers which should be cited based on the code that was used (alpha feature).
        settings: System settings.
        database: Database (or file base) connection.
        queue_adapter: Configuration for using remote resources.
    """

    @property
    def logger(self):
        return _logger

    @property
    def publications(self):
        return _publications

    @property
    def settings(self):
        return _settings

    @property
    def database(self):
        return _database

    @property
    def queue_adapter(self):
        return _queue_adapters.adapter

    def update(self, config_dict: Union[Dict, None] = None) -> None:
        """
        Re-reads the settings configuration, then reconstructs the queue adapter and reboots the database connection.

        Args:
            config_dict (dict): A new set of configuration parameters to use. (Default is None, which attempts to read
                the configuration from system environment xor configuration files.)
        """
        self.settings.update(user_dict=config_dict)
        _queue_adapters.update()
        self.database.update()


state = State()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Abstract classes for factories that instantiate other pyiron objects.
"""

from abc import ABC

__author__ = "Liam Huber"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.1"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Nov 20, 2020"


class PyironFactory(ABC):
    """
    A base class for factories, an abstraction layer which help facilitate tab-completion.
    """

    def __getitem__(self, key):
        return getattr(self, key)



"""
Data structure for versatile data handling.
"""

# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import copy
import json
import warnings
from collections.abc import Sequence, Set, Mapping, MutableMapping

import numpy as np
import pandas

from pyiron_base.storage.fileio import read, write
from pyiron_base.storage.hdfstub import HDFStub
from pyiron_base.interfaces.has_groups import HasGroups
from pyiron_base.interfaces.has_hdf import HasHDF

__author__ = "Marvin Poul"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.1"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "production"
__date__ = "Jun 17, 2020"


_internal_hdf_nodes = ["NAME", "TYPE", "OBJECT", "VERSION", "HDF_VERSION", "READ_ONLY"]


def _normalize(key):
    if isinstance(key, str):
        if key.isdecimal():
            return int(key)
        elif "/" in key:
            return tuple(key.split("/"))

    elif isinstance(key, tuple) and len(key) == 1:
        return _normalize(key[0])

    return key


class DataContainer(MutableMapping, HasGroups, HasHDF):
    """
    Mutable sequence with optional keys.

    If no argument is given, the constructor creates a new empty DataContainer.  If
    specified init maybe a Sequence, Set or Mapping and all recursive
    occurrences of these are also wrapped by DataContainer.

    >>> pl = DataContainer([3, 2, 1, 0])
    >>> pm = DataContainer({"foo": 24, "bar": 42})

    Access can be like a normal list with integers or optionally with strings
    as keys.

    >>> pl[0]
    3
    >>> pl[2]
    1
    >>> pm["foo"]
    24

    Keys do not have to be present for all elements.

    >>> pl2 = DataContainer([1,2])
    >>> pl2["end"] = 3
    >>> pl2
    DataContainer({0: 1, 1: 2, 'end': 3})

    It is also allowed to set an item one past the length of the DataContainer,
    this is then equivalent to appending that element.  This allows to use the
    update method also with other DataContainers

    >>> pl[len(pl)] = -1
    >>> pl
    DataContainer([3, 2, 1, 0, -1])
    >>> pl.pop(-1)
    -1

    Where strings are used they may also be used as attributes.  Getting keys
    which clash with methods of DataContainer must be done with item access, but
    setting them works without overwriting the instance methods, but is not
    recommended for readability.

    >>> pm.foo
    24
    >>> pm.tail = 23
    >>> pm
    DataContainer({'foo': 24, 'bar': 42, 'tail': 23})

    Keys and indices can be tuples to traverse nested DataContainers.

    >>> pn = DataContainer({"foo": {"bar": [4, 2]}})
    >>> pn["foo", "bar"]
    DataContainer([4, 2])
    >>> pn["foo", "bar", 0]
    4

    Using keys with "/" in them is equivalent to the above after splitting the
    key.

    >>> pn["foo/bar"] == pn["foo", "bar"]
    True
    >>> pn["foo/bar/0"] == pn["foo", "bar", 0]
    True

    To make that work strings that are only decimal digits are automatically
    converted to integers before accessing the list and keys are restricted to
    not only contain digits on initialization.

    >>> pl["0"] == pl[0]
    True
    >>> DataContainer({1: 42})
    Traceback (most recent call last):
        File "<stdin>", line 1, in <module>
        File "datacontainer.py", line 126, in __init__
            raise ValueError(
    ValueError: keys in initializer must not be int or str of decimal digits or in correct order, is 1

    When initializing from a dict, it may not have integers or decimal strings
    as keys unless they match their position in the insertion order.  This is
    to avoid ambiguities in the final order of the DataContainer.

    >>> DataContainer({0: "foo", 1: "bar", 2: 42})
    DataContainer(['foo', 'bar', 42])
    >>> DataContainer({0: "foo", 2: 42, 1: "bar"})
    Traceback (most recent call last):
        File "<stdin>", line 1, in <module>
        File "datacontainer.py", line 132, in __init__
            raise ValueError(
    ValueError: keys in initializer must not be int or str of decimal digits or in correct order, is 2


    Using keys is completely optional, DataContainer can always be treated as a
    list, with the exception that `iter()` iterates of the keys and indices.
    This is to correctly implement the MutableMapping protocol, to convert to a
    normal list and discard the keys use `values()`.

    >>> pm[0]
    24
    >>> pn["0/0/1"]
    2
    >>> list(pl)
    [0, 1, 2, 3]
    >>> list(pl.values())
    [3, 2, 1, 0]
    >>> list(pl.keys())
    [0, 1, 2, 3]

    Implements :class:`.HasGroups`.  Groups are nested data containers and nodes are everything else.

    >>> p = DataContainer({"a": 42, "b": [0, 1, 2]})
    >>> p.list_groups()
    ['b']
    >>> p.list_nodes()
    ['a']

    If instantiated with the argument `lazy=True`, data read from HDF5 later via :method:`.from_hdf` are not actually
    read, but only earmarked to be read later when actually accessed via :class:`.HDFStub`.  This is largely
    transparent, i.e. when accessing an earmarked value it will automatically be loaded and this loaded value is stored
    in container.  The only difference is in the string representation of the container, values not read yet appear as
    'HDFStub(...)' in the output.

    .. attention:: Subclasses beware!

        DataContainer require some careful treatment when creating subclasses.

        1. Since DataContainers are expected to recursively instantiate themselves subclasses need to keep their
        `__init__ compatible to the base class.  That means being able to be instantiated without arguments, if
        arguments are given the first one (or `init`) has to accept a Mapping or Iterable.  Additional arguments may be
        added, but must be after `init` and must have a default.

        2. Creating new instance attributes that don't live in the container itself is possible, but you need to use
        `object.__setattr__` the first time you define that attribute.  Afterwards using normal assignment syntax works.

        3. Subclasses should always be thought of as general data structures, if you want to subclass to have access to
        the HDF5 functionality or the way the DataContainer is shown in jupyter notebooks, but only have a fixed number
        of attributes it is better to create a new class that has an DataContainer as an attribute and dispatch to the
        :meth:`DataContainer.from_hdf`, :meth:`DataContainer.to_hdf` and :meth:`DataContainer._repr_json_`
        methods.
        4. To allow lazy loading sub classes must accept a `lazy` keyword argument and pass it to `super().__init__`.


    A few examples for subclasses

    >>> class ExtendedContainer(DataContainer):
    ...     def __init__(self, init=None, my_fancy_field=42, table_name=None, lazy=False):
    ...         super().__init__(init=init, table_name=table_name, lazy=lazy)
    ...         object.__setattr__(self, "my_fancy_field", my_fancy_field)

    After defining it once like this you can access my_fancy_field as a normal attribute, but it will not be stored in
    the container itself and will not be stored in HDF5.

    >>> e = ExtendedContainer({'foo': 1, 'bar': 5}, my_fancy_field=23)
    >>> e.my_fancy_field
    23
    >>> e
    ExtendedContainer({'foo': 1, 'bar': 5})
    >>> e.my_fancy_field = 42
    >>> e.my_fancy_field
    42
    >>> e
    ExtendedContainer({'foo': 1, 'bar': 5})

    Be aware the :class:`.DataContainer` and its subclasses are recursive data structures, i.e. your fancy attribute
    will be available also on sub groups.

    >>> g = e.create_group('sub')
    >>> g.fnord = 23
    >>> g.my_fancy_field
    42
    >>> e
    ExtendedContainer({'foo': 1, 'bar': 5, 'sub': ExtendedContainer({'fnord': 23})})

    For that reason most of time you'll actually want a class that uses a DataContainer for storage, but doesn't derive
    from it.

    >>> from pyiron_base.interfaces.object import HasStorage
    >>> class FancyClass(HasStorage):
    ...     def __init__(self, foo):
    ...         super().__init__()
    ...         self.storage.foo = foo
    ...
    ...     @property
    ...     def foo(self):
    ...         return self.storage.foo
    ...
    ...     @foo.setter
    ...     def foo(self, val):
    ...         self.storage.foo = val
    ...
    ...     def _repr_json_(self):
    ...         return self.storage._repr_json_()

    """

    __version__ = "0.1.0"
    __hdf_version__ = "0.2.0"

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        # setting these immediately after object creation ensures that they are
        # always defined and attribute access works even before __init__ is
        # called.  This is relevant on deepcopy & pickling.
        object.__setattr__(instance, "_store", [])
        object.__setattr__(instance, "_indices", {})
        object.__setattr__(instance, "table_name", None)
        object.__setattr__(instance, "_read_only", False)
        object.__setattr__(instance, "_lazy", False)

        return instance

    def __init__(self, init=None, table_name=None, lazy=False, wrap_blacklist=()):
        """
        Create new container.

        Args:
            init (Sequence, Mapping): initial data for the container, nested occurances of Sequence and Mapping are
                                      translated to nested containers
            table_name (str): default name of the data container in HDF5
            lazy (bool): if True, use :class:`.HDFStub` to load values lazily from HDF5
            wrap_blacklist (tuple of types): any values in `init` that are instances of the given types are *not*
                                             wrapped in :class:`.DataContainer`
        """
        self.table_name = table_name
        self._lazy = lazy
        if init is not None:
            self.update(init, wrap=True, blacklist=wrap_blacklist)

    def __len__(self):
        return len(self._store)

    def __iter__(self):
        reverse_indices = {i: k for k, i in self._indices.items()}

        for i in range(len(self)):
            yield reverse_indices.get(i, i)

    def __getitem__(self, key):
        key = _normalize(key)

        if isinstance(key, tuple):
            if key[0] == "..." and len(key) > 1:
                res = self.search(key[1], False)
                return res if (len(key) == 2) else res[key[2:]]
            return self[key[0]][key[1:]]

        elif isinstance(key, int):
            try:
                v = self._store[key]
                if not isinstance(v, HDFStub):
                    return v
                else:
                    v = self._store[key] = v.load()
                    return v
            except IndexError:
                raise IndexError("list index out of range") from None

        elif isinstance(key, str):
            try:
                v = self._store[self._indices[key]]
                if not isinstance(v, HDFStub):
                    return v
                else:
                    v = self._store[self._indices[key]] = v.load()
                    return v
            except KeyError:
                raise KeyError(repr(key)) from None

        else:
            raise ValueError("{} is not a valid key, must be str or int".format(key))

    def __setitem__(self, key, val):
        if self.read_only:
            self._read_only_error()

        key = _normalize(key)

        if isinstance(key, tuple):
            if key[0] == "..." and len(key) > 1:
                res = self._search_parent(key[1], False)
                res[key[1:]] = val
                return
            if key[0] not in self.keys():
                self[key[0]] = type(self)()
            self[key[0]][key[1:]] = val
        elif isinstance(key, int):
            if key < len(self):
                self._store[key] = val
            elif key == len(self):
                self.append(val)
            else:
                raise IndexError("index out of range")
        elif isinstance(key, str):
            if key not in self._indices:
                self._indices[key] = len(self._store)
                self._store.append(val)
            else:
                self._store[self._indices[key]] = val
        else:
            raise ValueError("{} is not a valid key, must be str or int".format(key))

    def __delitem__(self, key):
        if self.read_only:
            self._read_only_error()

        key = _normalize(key)

        if isinstance(key, tuple):
            if key[0] == "..." and len(key) > 1:
                res = self._search_parent(key[1], False)
                del res[key[1:]]
                return
            del self[key[0]][key[1:]]
        elif isinstance(key, (str, int)):
            if isinstance(key, str):
                idx = self._indices[key]
                del self._indices[key]
            else:
                idx = key

            del self._store[idx]

            for k, i in self._indices.items():
                if i > idx:
                    self._indices[k] = i - 1
        else:
            raise ValueError("{} is not a valid key, must be str or int".format(key))

    def __getattr__(self, name):
        # this is only called when python doesn't find name in the instance
        # or class variables, so we don't need to go through the same lengths
        # here as in __setattr__
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name) from None

    @classmethod
    def _is_class_var(cls, name):
        return any(name in c.__dict__ for c in cls.__mro__)

    def __setattr__(self, name, val):
        # Search instance variables (self.__dict___) and class variables
        # (self.__class__.__dict__ + iterating over mro to find variables on
        #  all ancestors) first before we assign the value into our container.
        # If we find name refers to a instance/class variable, we let
        # object.__setattr__ do all the work for us.
        if name in self.__dict__ or self._is_class_var(name):
            object.__setattr__(self, name, val)
        else:
            self[name] = val

    def __delattr__(self, name):
        # see __setattr__
        if name in self.__dict__ or self._is_class_var(name):
            object.__delattr__(self, name)
        else:
            del self[name]

    def __array__(self):
        """Return bare list of values to play nice with numpy."""
        return np.array(self._store)

    def __dir__(self):
        return set(super().__dir__() + list(self._indices.keys()))

    def __repr__(self):
        name = self.__class__.__name__
        if self.has_keys():
            # access _store and _indices directly to avoid forcing HDFStubs
            index2key = {v: k for k, v in self._indices.items()}
            return (
                name
                + "({"
                + ", ".join(
                    "{!r}: {!r}".format(index2key.get(i, i), self._store[i])
                    for i in range(len(self))
                )
                + "})"
            )
        else:
            return name + "([" + ", ".join("{!r}".format(v) for v in self._store) + "])"

    @property
    def read_only(self):
        """
        bool: if set, raise warning when attempts are made to modify the container
        """
        return self._read_only

    @read_only.setter
    def read_only(self, val):
        # can't mark a read-only list as writeable
        if self._read_only and not val:
            self._read_only_error()
        else:
            self._read_only = bool(val)

    @classmethod
    def _read_only_error(cls):
        warnings.warn(
            "The input in {} changed, while the state of the job was already "
            "finished.".format(cls.__name__)
        )

    def to_builtin(self, stringify=False):
        """
        Convert the container back to builtin dict's and list's recursively.

        Args:
            stringify (bool, optional): convert all non-recursive elements to str
        """

        if self.has_keys():
            dd = {}
            for k, v in self.items():
                # force all string keys in output to work with h5io (it
                # requires all string keys when storing as json), since
                # _normalize calls int() on all digit string keys this is
                # transparent for the rest of the module
                k = str(k)
                if isinstance(v, DataContainer):
                    dd[k] = v.to_builtin(stringify=stringify)
                else:
                    dd[k] = repr(v) if stringify else v

            return dd
        elif stringify:
            return list(
                v.to_builtin(stringify=stringify)
                if isinstance(v, DataContainer)
                else repr(v)
                for v in self.values()
            )
        else:
            return list(
                v.to_builtin(stringify=stringify) if isinstance(v, DataContainer) else v
                for v in self.values()
            )

    # allows "nice" displays in jupyter lab
    def _repr_json_(self):
        return self.to_builtin(stringify=True)

    # allows 'nice' display in notebooks
    def _repr_html_(self):
        name = self.__class__.__name__
        plain = f"{name}({json.dumps(self.to_builtin(stringify=True), indent=2, default=str)})"
        return "<pre>" + plain + "</pre>"

    def get(self, key, default=None, create=False):
        """
        If ``key`` exists, behave as generic, if not call create_group.

        Args:
            key (str):               key to search
            default (optional):      return this instead if nothing found
            create (bool, optional): create empty container at key if nothing found

        Raise:
            IndexError: if key is not in the container and neither ``default`` not
            ``create`` are given

        Returns:
            object: element at ``key`` or new empty subcontainer
        """
        if create and key not in self:
            return self.create_group(key)
        else:
            return super().get(key, default=default)

    def search(self, key, stop_on_first_hit=True):
        """
        Search for ``key`` in the Container hierarchy.

        This should be used if there is only one such item in the hierarchy.

        If stop_on_first_hit is True the first item found is taken.
        Otherwise, a ValueError is raised if the key appears multiple times.

        Args:
            key (str):                the key to look for
            stop_on_first_hit (bool): whether to stop on the first hit

        Raise:
            TypeError:  if key is not str
            KeyError:   if key is not found
            ValueError: if stop_on_first_hit is False and key is found twice

        Returns:
            object: element at ``key``
        """

        if not isinstance(key, str):
            raise TypeError("Cannot search for non-string key.")

        parent = self._search_parent(key, stop_on_first_hit)
        if parent is None:
            raise KeyError("Could not find any element '" + key + "' in tree.")

        return parent[key]

    def _search_parent(self, key, stop_on_first_hit=True):
        """
        Search for container in hierarchy which has ``key``

        This should be used if there is only one such item in the hierarchy.
        If stop_on_first_hit is True the first item found is taken.
        Otherwise, a ValueError is raised if the key appears multiple times.

        Args:
            key (str):                the key to look for
            stop_on_first_hit (bool): what to do if key is found
                                      True  => return
                                      False => continue to check that it is
                                               the only hit
        Raise:
            ValueError: if key is found twice and stop_on_first_hit is False

        Returns:
            DataContainer: container that has ``key``
        """
        # search within current level
        if key in self:
            if stop_on_first_hit:
                return self
            else:
                first_hit = self
        else:
            first_hit = None

        # descend into subgroups
        for it in self.groups():
            hit = self[it]._search_parent(key, stop_on_first_hit)
            if isinstance(hit, DataContainer):
                if stop_on_first_hit:
                    return hit
                else:
                    if isinstance(first_hit, DataContainer):
                        raise ValueError("'" + key + "' exists more than once!")
                first_hit = hit
        return first_hit

    @classmethod
    def _wrap_val(cls, val, blacklist):
        if isinstance(val, (Sequence, Set, Mapping)) and not isinstance(val, blacklist):
            return cls(val, wrap_blacklist=blacklist)
        else:
            return val

    def update(self, init, wrap=False, blacklist=(), **kwargs):
        """
        Add all elements or key-value pairs from init to this container.  If wrap is
        not given, behaves as the generic method.

        Args:
            init (Sequence, Set, Mapping): container to draw new elements from
            wrap (bool): if True wrap all encountered Sequences and Mappings in
                         :class:`.DataContainer` recursively
            blacklist (list of types): when `wrap` is True, don't wrap these types even if they're instances of Sequence
                                       or Mapping
            **kwargs: update from this mapping as well
        """
        if str not in blacklist:
            blacklist += (str,)
        if wrap and (isinstance(wrap, bool) or not isinstance(init, blacklist)):
            if isinstance(init, (Sequence, Set)):
                for v in init:
                    self.append(self._wrap_val(v, blacklist))

            elif isinstance(init, Mapping):
                for i, (k, v) in enumerate(init.items()):
                    k = _normalize(k)
                    v = self._wrap_val(v, blacklist)
                    if isinstance(k, int):
                        if k == i:
                            self.append(v)
                        else:
                            raise ValueError(
                                "keys in initializer must not be int or str of "
                                "decimal digits or in correct order, "
                                "is {!r}".format(k)
                            )
                    else:
                        self[k] = v
            else:
                ValueError("init must be Sequence, Set or Mapping")

            for k in kwargs:
                self[k] = self._wrap_val(kwargs[k], blacklist)
        else:
            super().update(init, **kwargs)

    def append(self, val):
        """
        Add new value to the container without a key.

        Args:
            val: new element
        """
        self._store.append(val)

    def extend(self, vals):
        """
        Append vals to the end of this DataContainer.

        Args:
            vals (Sequence): any python sequence to draw new elements from
        """

        for v in vals:
            self.append(v)

    def insert(self, index, val, key=None):
        """
        Add a new element to the container at the specified position, with an optional
        key.  If the key is already in the container it will be updated to point to
        the new element at the new index.  If index is larger than container, append
        instead.

        Args:
            index (int):            place val after this element
            val:                    new element to add
            key (str, optional):    optional key to mark the new element
        """
        if key is not None:
            for k, i in self._indices.items():
                if i >= index:
                    self._indices[k] = i + 1
            self._indices[key] = index

        self._store.insert(index, val)

    def mark(self, index, key):
        """
        Add a key to an existing item at index.  If key already exists, it is
        overwritten.

        Args:
            index (int):    index of the existing element to mark
            key (str):      key for the existing element

        Raises:
            IndexError: if index > len(self)

        >>> pl = DataContainer([42])
        >>> pl.mark(0, "head")
        >>> pl.head == 42
        True
        """
        if index >= len(self):
            raise IndexError("list index out of range")

        reverse_indices = {i: k for k, i in self._indices.items()}
        if index in reverse_indices:
            del self._indices[reverse_indices[index]]

        self._indices[key] = index

    def clear(self):
        """
        Remove all items from DataContainer.
        """
        self._store.clear()
        self._indices.clear()

    def create_group(self, name):
        """
        Add a new empty subcontainer under the given key.

        Args:
            name (str): key under which to store the new subcontainer in this container

        Returns:
            DataContainer: the newly created subcontainer

        Raises:
            ValueError: name already exists in container and is not a sub container

        >>> pl = DataContainer({})
        >>> pl.create_group("group_name")
        DataContainer([])
        >>> list(pl.group_name)
        []
        """
        if name not in self:
            self[name] = self.__class__()
            return self[name]
        else:
            v = self[name]
            if isinstance(v, self.__class__):
                return v
            else:
                raise ValueError(f"'{name}' already exists in DataContainer.")

    def has_keys(self):
        """
        Check if the container has keys set or not.

        Returns:
            bool: True if there is at least one key set
        """
        return bool(self._indices)

    def __copy__(self):
        # by default copy.copy will use the same objects for _store and
        # _indices, which would cause the copied and the copiee to have the
        # same underlying data storage, so instead we have to do a shallow copy
        # of those manually
        copiee = type(self)()
        copiee._store = copy.copy(self._store)
        copiee._indices = copy.copy(self._indices)
        copiee.table_name = self.table_name
        return copiee

    def copy(self):
        """
        Returns deep copy of it self.  A shallow copy can be obtained via the
        copy module.

        Returns:
            DataContainer: deep copy of itself

        >>> pl = DataContainer([[1,2,3]])
        >>> pl.copy() == pl
        True
        >>> pl.copy() is pl
        False
        >>> all(a is not b for a, b in zip(pl.copy().values(), pl.values()))
        True
        """
        self._force_load()
        return copy.deepcopy(self)

    def _get_hdf_group_name(self):
        return self.table_name

    def _to_hdf(self, hdf):
        hdf["READ_ONLY"] = self.read_only
        written_keys = _internal_hdf_nodes.copy()
        for i, (k, v) in enumerate(self.items()):
            if isinstance(k, str) and "__index_" in k:
                raise ValueError("Key {} clashes with internal use!".format(k))

            k = "{}__index_{}".format(k if isinstance(k, str) else "", i)
            written_keys.append(k)

            # pandas objects also have a to_hdf method that is entirely unrelated to ours
            if hasattr(v, "to_hdf") and not isinstance(
                v, (pandas.DataFrame, pandas.Series)
            ):
                # if v will be written as a group, but a node of the same name k exists already in the file, h5py will
                # complain, so delete it first
                if k in hdf.list_nodes():
                    del hdf[k]
                v.to_hdf(hdf=hdf, group_name=k)
            else:
                # if the value doesn't know how to serialize itself, assume
                # that h5py knows how to
                try:
                    hdf[k] = v
                except TypeError:
                    raise TypeError(
                        "Error saving {} (key {}): DataContainer doesn't support saving elements "
                        'of type "{}" to HDF!'.format(v, k, type(v))
                    ) from None
        for n in hdf.list_nodes():
            if n not in written_keys:
                del hdf[n]
        for g in hdf.list_groups():
            if g not in written_keys:
                del hdf[g]

    def _from_hdf(self, hdf, version=None):
        self.clear()

        if version == "0.1.0":
            self.update(hdf["data"], wrap=True)
            self.read_only = bool(hdf.get("read_only", False))
        else:

            def normalize_key(name):
                # split a dataset/group name into the position in the list and
                # the key
                if "__index_" in name:
                    k, i = name.split("__index_", maxsplit=1)
                else:
                    k = name
                    i = -1
                i = int(i)
                if k == "":
                    return i, i
                else:
                    return i, k

            items = []
            for n in hdf.list_nodes():
                if n in _internal_hdf_nodes:
                    continue
                items.append(
                    (*normalize_key(n), hdf[n] if not self._lazy else HDFStub(hdf, n))
                )
            for g in hdf.list_groups():
                items.append(
                    (
                        *normalize_key(g),
                        hdf[g].to_object() if not self._lazy else HDFStub(hdf, g),
                    )
                )

            for _, k, v in sorted(items, key=lambda x: x[0]):
                self[k] = v

            self.read_only = bool(hdf.get("READ_ONLY", False))

    def nodes(self):
        """
        Iterator over keys to terminal nodes.

        Returns:
            :class:`list`: list of keys to normal values.
        """
        for k, v in self.items():
            if not isinstance(v, DataContainer):
                yield k

    def _list_nodes(self):
        return list(self.nodes())

    def groups(self):
        """
        Iterate over keys to nested containers.

        Returns:
            :class:`list`: list of all keys to elements of :class:`DataContainer`.
        """
        for k, v in self.items():
            if isinstance(v, DataContainer):
                yield k

    def _list_groups(self):
        return list(self.groups())

    def read(self, file_name, wrap=True):
        """
        Parse file as dictionary and add its keys to this container.

        For supported file types, see :func:`.fileio.read`.

        Errors during reading of the files generate a warning, but leave the container unchanged.

        Args:
            file_name(str): path to the input file
            wrap(bool), if set to true will wrap the inputed data itself as a datacontainer inside the datacontainer

        Raises:
            :class:`ValueError`: if file extension doesn't match one of the supported ones
        """
        self.update(read(file_name), wrap=wrap)

    def write(self, file_name):
        """
        Writes the DataContainer to a text file.

        For supported file types, see :func:`.fileio.write`.

        Args:
            file_name(str): the name of the file to be writen to.
        """
        write(self.to_builtin(), file_name)

    def _force_load(self, recursive=True):
        """
        Load all HDFStubs present in the data container.

        Args:
            recursive (bool): force also nested data containers, default True
        """
        if not self._lazy and not recursive:
            return

        # values are loaded from HDF once they are accessed via __getitem__, which is implicitly called by values()
        for v in self.values():
            if recursive and isinstance(v, DataContainer):
                v._force_load()

    def __init_subclass__(cls):
        # called whenever a subclass of DataContainer is defined, then register all subclasses with the same function
        # that the DataContainer is registered
        HDFStub.register(cls, lambda h, g: h[g].to_object(lazy=True))


HDFStub.register(DataContainer, lambda h, g: h[g].to_object(lazy=True))



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Load input parameters for jupyter notebooks from external HDF5 or JSON file
"""

import json
from pathlib2 import Path
import warnings
from pyiron_base.storage.hdfio import FileHDFio
from pyiron_base.storage.datacontainer import DataContainer

__author__ = "Osamu Waseda"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2019"


class Notebook(object):
    """
    class for pyiron notebook objects

    # TODO: Extract JSON functionality over to Project.data.read/write and remove this file.
    """

    @staticmethod
    def get_custom_dict():
        return load()

    @staticmethod
    def store_custom_output_dict(output_dict):
        dump(output_dict=output_dict)


def load():
    folder = Path(".").cwd().parts[-1]
    project_folder = Path(".").cwd().parents[1]
    hdf_file = project_folder / folder
    hdf_file = str(hdf_file).replace("\\", "/") + ".h5"
    if Path(hdf_file).exists():
        obj = DataContainer()
        obj.from_hdf(hdf=FileHDFio(hdf_file), group_name=folder + "/input/custom_dict")
        obj["project_dir"] = str(project_folder)
        return obj
    elif Path("input.json").exists():
        with open("input.json") as f:
            return json.load(f)
    else:
        warnings.warn("{} not found".format(hdf_file))
        return None


def dump(output_dict):
    folder = Path(".").cwd().parts[-1]
    hdf_file = Path(".").cwd().parents[1] / folder
    hdf_file = str(hdf_file) + ".h5"
    hdf = FileHDFio(hdf_file)
    hdf[folder].create_group("output")
    obj = DataContainer(output_dict)
    obj.to_hdf(hdf[folder + "/output"])



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
The core class in pyiron, linking python to the database to file storage.
"""

from abc import ABC
from pyiron_base.storage.datacontainer import DataContainer
from pyiron_base.interfaces.has_hdf import HasHDF
from pyiron_base.storage.hdfio import ProjectHDFio
from pyiron_base.state import state

__author__ = "Liam Huber"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Mar 23, 2021"


class HasStorage(HasHDF, ABC):
    """
    A base class for objects that use HDF5 data serialization via the `DataContainer` class.

    Unless you know what you are doing sub classes should pass the `group_name` argument to :meth:`~.__init__` or
    override :meth:`~.get_hdf_group_name()` to force a default name for the HDF group the object should write itself to.
    """

    def __init__(self, *args, group_name=None, **kwargs):
        """

        Args:
            group_name (str): default name of the HDF group where the whole object should be written to.
        """
        super().__init__(*args, **kwargs)
        self._storage = DataContainer(table_name="storage")
        self._group_name = group_name

    @property
    def storage(self) -> DataContainer:
        return self._storage

    def _get_hdf_group_name(self) -> str:
        return self._group_name

    def _to_hdf(self, hdf: ProjectHDFio):
        self.storage.to_hdf(hdf=hdf)

    def _from_hdf(self, hdf: ProjectHDFio, version: str = None):
        self.storage.from_hdf(hdf=hdf)


class HasDatabase(ABC):
    """
    A base class for objects that are registered in pyiron's database

    # TODO: Flesh this class out so it actually gives us a good link to the database!
    """

    def __init__(self, *args, **kwargs):
        if not state.database.database_is_disabled:
            state.database.open_connection()
            self._database = state.database.database
        else:
            raise NotImplementedError("WIP. For now only allowed with a database")

    @property
    def database(self):
        return self._database

    def save(self):
        raise NotImplementedError(
            "WIP. Saving should make sure you're registered with the database."
        )


class PyironObject(HasStorage, HasDatabase, ABC):
    """
    The fundamental pyiron object bringing together python objects, database identification, and data serialziation.
    """

    def __init__(self, *args, **kwargs):
        HasStorage.__init__(self)
        HasDatabase.__init__(self)



from pyiron_base._tests import TestWithProject
import numpy as np
import timeit

class TestFileHDFio(TestWithProject):
    def test_ragged(self):
        """Storing normal arrays should be faster than storing ragged arrays."""
        hdf = self.project.create_hdf(self.project.path, "ragged")
        regular = np.random.rand(10, 100, 100)
        ragged  = [np.random.rand(100 + i, 100) for i in range(-5, 5)]

        time_regular = timeit.timeit("hdf['array'] = array", number=50, globals={'hdf': hdf, 'array': regular})
        time_ragged = timeit.timeit("hdf['array'] = array", number=50, globals={'hdf': hdf, 'array': ragged})
        self.assertGreater(time_ragged, time_regular,
                           "Storing an regular array is not faster than a ragged one!")



# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
CLI for various pyiron utilities.
"""

import argparse
import os
import warnings

from pyiron_base.cli import ls, rm, install, reloadfile, wrapper

__author__ = "Marvin Poul"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "development"
__date__ = "26 Jun, 2020"

cli_modules = {
    "ls": ls,
    "rm": rm,
    "install": install,
    "reloadfile": reloadfile,
    "wrapper": wrapper,
}


def main():
    parser = argparse.ArgumentParser(prog="pyiron", description=__doc__)
    parser.add_argument(
        "-d", "--dirty", action="store_true", help="do not remove pyiron log files"
    )
    subs = parser.add_subparsers()

    parser.set_defaults(cli=lambda _: parser.error("no sub command given"))

    for name, mod in cli_modules.items():
        try:
            sub_parser = subs.add_parser(
                name,
                help=mod.__doc__,
                description=mod.__doc__,
                epilog=getattr(mod, "epilog", None),
                formatter_class=getattr(mod, "formatter", argparse.HelpFormatter),
            )
            sub_parser.set_defaults(cli=mod.main)
            mod.register(sub_parser)
        except AttributeError:
            warnings.warn(
                "module '{}' does not define main or register " "function, ignoring"
            )

    args = parser.parse_args()
    args.cli(args)

    if not args.dirty and os.path.exists("pyiron.log"):
        os.remove("pyiron.log")



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Classes defining the shape of pyiron's database tables.
"""

from sqlalchemy import (
    Column,
    DateTime,
    Float,
    Integer,
    String,
    Table,
)

__author__ = "Murat Han Celik, Liam Huber"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH"
    " - Computational Materials Design (CM) Department"
)
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Sep, 2021"


def get_historical_table(table_name, metadata, extend_existing=True):
    """The historical table."""
    return Table(
        table_name,
        metadata,
        Column("id", Integer, primary_key=True, autoincrement=True),
        Column("parentid", Integer),
        Column("masterid", Integer),
        Column("projectpath", String(50)),
        Column("project", String(255)),
        Column("job", String(50)),
        Column("subjob", String(255)),
        Column("chemicalformula", String(50)),
        Column("status", String(20)),
        Column("hamilton", String(20)),
        Column("hamversion", String(50)),
        Column("username", String(20)),
        Column("computer", String(100)),
        Column("timestart", DateTime),
        Column("timestop", DateTime),
        Column("totalcputime", Float),
        extend_existing=extend_existing,
    )
